/**
 * Copyright (c) 2012 EBMWebsourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.probes.impl;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import org.ow2.petals.probes.api.enums.ExecutionStatus;
import org.ow2.petals.probes.api.exceptions.NoResponseTimeException;
import org.ow2.petals.probes.api.exceptions.ResponseTimeCollectionStoppedException;
import org.ow2.petals.probes.api.exceptions.StartDateItemLostException;
import org.ow2.petals.probes.api.exceptions.StartDateItemUnknownException;
import org.ow2.petals.probes.api.key.ExecutionStatusProbeKey;
import org.ow2.petals.probes.api.key.ProbeKey;
import org.ow2.petals.probes.api.probes.KeyedStartDateItem;
import org.ow2.petals.probes.api.probes.ResponseTimeRelativeValues;
import org.ow2.petals.probes.key.ExecutionStatusProbeKeyImpl;

/**
 * <p>
 * Response time sample collecting response times during its sampling duration.
 * When collection ends, relative response time values are computed.
 * </p>
 * <p>
 * Relative response time values are defined by
 * {@link ResponseTimeRelativeValues}
 * </p>
 * 
 * @author Christophe DENEUX (PetalsLink)
 * 
 * @param <K>
 *            Type of the key grouping response time values, implementing
 *            {@link ProbeKey}.
 */
public class KeyedResponseTimesSample<K extends ProbeKey> {

    /**
     * <p>
     * List of start date of pending executions. When a execution ends, the end
     * time is available and the response is moved to {@link #responseTimes}.
     * </p>
     */
    private final List<KeyedStartDateItem<K>> startDates;

    private final ReadWriteLock startDatesLock = new ReentrantReadWriteLock();

    /**
     * <p>
     * List of response times (of ended executions) by key of the current
     * sample. This list is filled when the sample runs. This list is clear when
     * the sample ends. The content of this list is used to update the max
     * response time values.
     * </p>
     * <p>
     * <u>Note:</u> {@link LinkedList} is used in definition instead of
     * {@link List} because of performance reasons, access to a linked list
     * elements is faster with the {@link LinkedList} API than using the
     * {@link List} API.
     * </p>
     */
    private final ConcurrentMap<ExecutionStatusProbeKey<K>, LinkedList<Long>> responseTimes = new ConcurrentHashMap<ExecutionStatusProbeKey<K>, LinkedList<Long>>();

    /**
     * List of response time values (max, avg, min, percentile, ...) of the
     * current sample. It is computed from response times available into
     * {@link #responseTimes} and {@link #startDates} when the sample is
     * stopped.
     */
    private final Map<ExecutionStatusProbeKey<K>, ResponseTimeRelativeValues> responseTimeValues = new HashMap<ExecutionStatusProbeKey<K>, ResponseTimeRelativeValues>();

    /**
     * <p>
     * Flag to control the sample lifecycle. When the sample ends, no more
     * response times is collected and response time values (max, avg, min,
     * percentiles, ...) are computed.
     * </p>
     */
    private volatile boolean isEchantillonRunning = true;

    /**
     * Lock object to synchronize the collection of response times
     */
    private final Object stopCollectionLock = new Object();

    /**
     * <p>
     * Creates a new sample from scratch. It <b>MUST</b> be the constructor to
     * use to create the <b>FIRST</b> sample.
     * </p>
     */
    public KeyedResponseTimesSample() {
        this.startDates = new LinkedList<KeyedStartDateItem<K>>();
    }

    /**
     * <p>
     * Creates a new sample from the current one:
     * <ul>
     * <li>the start dates are transfered,</li>
     * <li>the response times are not transfered,</li>
     * <li>the sample status is reset.</li>
     * </ul>
     * </p>
     */
    public KeyedResponseTimesSample(final KeyedResponseTimesSample<K> currentRelativeResponseTimes) {
        // it shouldn't be necessary to make a copy of the startDates because once a new sample is created based
        // on a previous one, the previous one is not updated anymore...
        this.startDates = currentRelativeResponseTimes.startDates;
    }

    /**
     * Add a new start date of a pending execution.
     * 
     * @param key
     *            The key associated to the current execution and grouping
     *            response time values
     * @param startTime
     *            The start time (timestamp)
     * @return A {@link KeyedStartDateItem} created for this new execution.
     * @throws ResponseTimeCollectionStoppedException
     *             The sample collecting response times is stopped.
     * @throws StartDateItemLostException
     *             Unable to add the new start date to the list of start dates.
     *             This exception should not occur.
     */
    public KeyedStartDateItem<K> addStartDate(final K key, final long startTime)
            throws ResponseTimeCollectionStoppedException, StartDateItemLostException {
        synchronized (this.stopCollectionLock) {
            if (this.isEchantillonRunning) {
                final KeyedStartDateItem<K> startDateItem = new KeyedStartDateItem<K>(key, startTime);
                startDatesLock.writeLock().lock();
                try {
                    if (this.startDates.add(startDateItem)) {
                        return startDateItem;
                    } else {
                        throw new StartDateItemLostException();
                    }
                } finally {
                    startDatesLock.writeLock().unlock();
                }
            } else {
                throw new ResponseTimeCollectionStoppedException();
            }
        }
    }

    /**
     * The execution associated to <code>startDateItem</code> ends.
     * 
     * @param startDateItem
     *            The {@link KeyedStartDateItem} created by
     *            {@link #addStartDate(ProbeKey, long)} for the current
     *            execution.
     * @param endTime
     *            The time where the execution ends (timestamp)
     * @param executionStatus
     *            The execution status with which the execution ends
     * @throws ResponseTimeCollectionStoppedException
     *             The sample collecting response times is stopped. This
     *             execution should not occur.
     * @throws StartDateItemUnknownException
     *             <code>startDateItem</code> not found in the list of start
     *             dates. This exception should not occur.
     */
    public void endStartDate(final KeyedStartDateItem<K> startDateItem, final long endTime,
            final ExecutionStatus executionStatus) throws ResponseTimeCollectionStoppedException,
            StartDateItemUnknownException {
        synchronized (this.stopCollectionLock) {
            if (this.isEchantillonRunning) {
                startDatesLock.writeLock().lock();
                try {
                    if (!this.startDates.remove(startDateItem)) {
                        throw new StartDateItemUnknownException(startDateItem, executionStatus);
                    }
                } finally {
                    startDatesLock.writeLock().unlock();
                }
                final K startDateKey = startDateItem.key;
                final ExecutionStatusProbeKey<K> responseTimeKey = new ExecutionStatusProbeKeyImpl<K>(
                        startDateKey, executionStatus);

                // Add the response time to its list
                final Long responseTime = endTime - startDateItem.startDate;
                final LinkedList<Long> emptyResponseTimeList = new LinkedList<Long>();
                final List<Long> responseTimeList = this.responseTimes.putIfAbsent(responseTimeKey,
                        emptyResponseTimeList);
                if (responseTimeList != null) {
                    responseTimeList.add(responseTime);
                } else {
                    emptyResponseTimeList.add(responseTime);
                }
            } else {
                throw new ResponseTimeCollectionStoppedException();
            }
        }
    }

    /**
     * Stops the sample to collect response times, and compute response time
     * values.
     */
    public void stopCollecting() {
        synchronized (this.stopCollectionLock) {

            // Stop collecting
            this.isEchantillonRunning = false;

            // Timestamp of the collecting stop (needed to compute elapsed time
            // of pending execution)
            final long stopTimestamp = System.currentTimeMillis();

            // Compute response time values
            for (final Map.Entry<ExecutionStatusProbeKey<K>, LinkedList<Long>> entry : this.responseTimes.entrySet()) {
                final LinkedList<Long> responseTimes = entry.getValue();
                if (!responseTimes.isEmpty()) {
                    try {
                        this.responseTimeValues.put(entry.getKey(), new ResponseTimeRelativeValues(responseTimes));
                    } catch (final NoResponseTimeException e) {
                        // This exception should not occur because of the
                        // previous check about number of response times
                    }
                } else {
                    // No response time collected for the key.
                }
            }

            // Sort the list of start dates by key
            final Map<K, LinkedList<Long>> startDatesByKey = new HashMap<>();
            startDatesLock.readLock().lock();
            try {
                for (final KeyedStartDateItem<K> startDate : this.startDates) {
                    final long responseTime = stopTimestamp - startDate.startDate;

                    final LinkedList<Long> startDateList = startDatesByKey.get(startDate.key);
                    if (startDateList != null) {
                        startDateList.add(responseTime);
                    } else {
                        final LinkedList<Long> newList = new LinkedList<Long>();
                        newList.add(responseTime);
                        startDatesByKey.put(startDate.key, newList);
                    }
                }
            } finally {
                startDatesLock.readLock().unlock();
            }
            // Add the pending execution to the response time values.
            for (final Map.Entry<K, LinkedList<Long>> entry : startDatesByKey.entrySet()) {
                final LinkedList<Long> responseTimes = entry.getValue();
                if (responseTimes.size() != 0) {
                    try {
                        this.responseTimeValues.put(
                                new ExecutionStatusProbeKeyImpl<K>(entry.getKey(), ExecutionStatus.PENDING),
                                new ResponseTimeRelativeValues(responseTimes));
                    } catch (final NoResponseTimeException e) {
                        // This exception should not occur because of the
                        // previous check about number of response times
                    }
                } else {
                    // No response time collected for the key.
                }
            }
        }
    }

    /**
     * @return The relative response time values
     */
    public Map<ExecutionStatusProbeKey<K>, ResponseTimeRelativeValues> getResponseTimeValues() {
        synchronized (this.stopCollectionLock) {
            // return a copy, if not it could be read or modified when we don't want to!
            return new HashMap<>(this.responseTimeValues);
        }
    }
}
